{-# LANGUAGE MultiParamTypeClasses, FlexibleInstances, ExistentialQuantification #-}

module OT where
import Control.Applicative

class Operation o where
  compose :: o -> o -> o
  transform :: o -> o -> (o,o)

class Operation o => Document d o where
  noop :: d -> o
  applyOp :: d -> o -> d

---

data TextAction = Retain
                | Insert Char
                | Delete deriving Show

newtype TextOperation = TextOperation [TextAction] deriving Show

instance Operation TextOperation where
  -- applyOp (compose a b) d === applyOp b (applyOp a d)
  compose (TextOperation a) (TextOperation b) = TextOperation $ compose' a b where
    compose' [] [] = []
    compose' (Delete : as) bs            = Delete : compose' as bs
    compose' as (Insert c : bs)          = Insert c : compose' as bs
    compose' (Retain : as) (Retain:bs)   = Retain : compose' as bs
    compose' (Retain : as) (Delete:bs)   = Delete : compose' as bs
    compose' (Insert c : as) (Retain:bs) = Insert c : compose' as bs
    compose' (Insert _ : as) (Delete:bs) = compose' as bs
    compose' _ _ = error "operations cannot be composed"
  -- transform a b = (a',b') --> compose a b' === compose b a'
  transform (TextOperation a) (TextOperation b) = pair $ transform' a b where
    pair (as',bs') = (TextOperation as', TextOperation bs')
    addTransform as bs a b = let (as',bs') = transform' as bs in (a ++ as', b ++ bs')
    transform' [] [] = ([],[])
    transform' (Insert c : as) bs          = addTransform as bs [Insert c] [Retain]
    transform' as (Insert c : bs)          = addTransform as bs [Retain]   [Insert c]
    transform' (Retain : as) (Retain : bs) = addTransform as bs [Retain]   [Retain]
    transform' (Delete : as) (Delete : bs) = addTransform as bs []         []
    transform' (Retain : as) (Delete : bs) = addTransform as bs []         [Delete]
    transform' (Delete : as) (Retain : bs) = addTransform as bs [Delete]   []
    transform' _ _ = error "operations cannot be transformed"

instance Document String TextOperation where
  noop = TextOperation . map (\_ -> Retain)
  applyOp d (TextOperation as) = apply' as d where
    apply' [] [] = []
    apply' (Retain : as) (c:cs) = c : apply' as cs
    apply' (Insert c : as) d = c : apply' as d
    apply' (Delete : as) (_:cs) = apply' as cs

---

data Client o d = (Operation o, Document d o) => Client {
  clientRevision :: Integer,
  pending :: Maybe o,
  buffer :: Maybe o
}

newClient :: (Operation o, Document d o) => Client o d
newClient = Client 0 Nothing Nothing

localEdit :: Client o d -> o -> (Bool, Client o d)
localEdit (Client rev pending buffer) operation = (null pending, Client rev newPending newBuffer) where
    newPending = pending <|> (Just operation)
    newBuffer  = fmap (`compose` operation) buffer <|> fmap (const operation) pending

remoteEdit :: Client o d -> o -> (o, Client o d)
remoteEdit (Client rev pending buffer) o = case (pending,buffer) of
    (Nothing,Nothing) -> (o, Client (rev + 1) Nothing Nothing)
    (Just p,Nothing)  -> let (p',o') = transform p o in
                             (o',Client (rev + 1) (Just p') Nothing)
    (Just p,Just b)   -> let (p',o') = transform p o
                             (b',o'') = transform b o' in
                             (o'',Client (rev + 1) (Just p') (Just b'))

ack :: Client o d -> (Maybe o, Client o d)
ack (Client rev pending buffer) = (buffer, Client (rev + 1) buffer Nothing)

isSynchronized :: Client o d -> Bool
isSynchronized (Client rev pending buffer) = null pending

---

data Server o d = (Operation o, Document d o) => Server {
    initialState :: d,
    history :: [o]
}

serverState :: Server o d -> d
serverState (Server start ops) = foldl applyOp start ops

serverStateOp :: Server o d -> o
serverStateOp (Server start ops) = foldl compose (noop start) ops

serverRevision :: Server o d -> Int
serverRevision (Server start ops) = length ops

appendOperation :: Server o d -> o -> Int -> (o,Server o d)
appendOperation (Server start ops) op rev = (op', Server start (ops ++ [op'])) where
  op' = foldl (\a b -> fst $ transform a b) op $ drop rev ops